from trame.app import get_server, jupyter
from vera_core.app import engine, ui

from .core.vera_out_file import VeraOutFile


def show(data=None, server=None, **kwargs):
    """Run and display the trame application in jupyter's event loop

    The kwargs are forwarded to IPython.display.IFrame()
    """
    if server is None:
        server = get_server()

    if isinstance(server, str):
        server = get_server(server)

    # Disable logging
    import logging

    engine_logger = logging.getLogger("vera_core.app.engine")
    engine_logger.setLevel(logging.WARNING)

    vera_out_file = VeraOutFile(data)

    # Initialize app
    engine.initialize(server)
    ui.initialize(server, vera_out_file)

    # Show as cell result
    jupyter.show(server, **kwargs)


def jupyter_proxy_info():
    """Get the config to run the trame application via jupyter's server proxy

    This is provided to the `jupyter_serverproxy_servers` entrypoint, and the
    jupyter server proxy will use it to start the application as a separate
    process.
    """
    return {
        "command": ["vera-core", "-p", "0", "--server"],
    }



from trame_client.widgets.core import AbstractElement
from .. import module


class HtmlElement(AbstractElement):
    def __init__(self, _elem_name, children=None, **kwargs):
        super().__init__(_elem_name, children, **kwargs)
        if self.server:
            self.server.enable_module(module)


class AssemblyView(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "vera-assembly-view",
            **kwargs,
        )
        self._attr_names += [
            "value",
            ("selected_i", "selectedI"),
            ("selected_j", "selectedJ"),
            ("color_preset", "colorPreset"),
            ("color_range", "colorRange"),
            ("active_style", ":activeStyle"),
        ]
        self._event_names += [
            "click",
        ]


class CoreView(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "vera-core-view",
            **kwargs,
        )
        self._attr_names += [
            "value",
            ("selected_i", "selectedI"),
            ("selected_j", "selectedJ"),
            ("color_preset", "colorPreset"),
            ("color_range", "colorRange"),
            ("active_style", ":activeStyle"),
            ("x_labels", "xLabels"),
            ("y_labels", "yLabels"),
            "scaling",
        ]
        self._event_names += [
            "click",
        ]


class AxialView(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "vera-axial-view",
            **kwargs,
        )
        self._attr_names += [
            "value",
            ("selected_i", "selectedI"),
            ("selected_j", "selectedJ"),
            ("color_preset", "colorPreset"),
            ("color_range", "colorRange"),
            ("active_style", ":activeStyle"),
            ("x_labels", "xLabels"),
            ("y_labels", "yLabels"),
            ("x_scale", "xScale"),
            ("y_scale", "yScale"),
            ("x_sizes", "xSizes"),
            ("y_sizes", "ySizes"),
        ]
        self._event_names += [
            "click",
        ]


class ColorMapEditor(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "vera-color-map-editor",
            **kwargs,
        )
        self._attr_names += [
            "value",
            ("color_preset", "colorPreset"),
        ]
        self._event_names += [
            "input",
        ]



from trame.ui.html import DivLayout

OPTION = {
    "name": "empty",
    "label": "Undefined content",
    "icon": "mdi-help-circle-outline",
}


def initialize(server):
    state = server.state

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    with DivLayout(server, template_name="empty") as layout:
        layout.root.add_child("Some empty content...")



from functools import partial
import os

from trame.app import get_server, dev

from . import ui
from .core.vera_out_file import VeraOutFile

# The user can set this via an environment variable
DATA_PATH_ENV_NAME = "VERA_CORE_DATA_PATH"


def _reload(vera_out_file):
    server = get_server()
    dev.reload(ui)
    ui.initialize(server, vera_out_file)


def main(server=None, **kwargs):

    # Get or create server
    if server is None:
        server = get_server()

    if isinstance(server, str):
        server = get_server(server)

    data_kwargs = {
        "help": "Data file to load",
        "dest": "data_file",
    }

    default = os.getenv(DATA_PATH_ENV_NAME)
    if default is not None:
        # If the environment variable has been provided, use that for the default
        data_kwargs["default"] = default
    else:
        # Otherwise, the CLI argument is required
        data_kwargs["required"] = True

    server.cli.add_argument("--data", **data_kwargs)
    args, _ = server.cli.parse_known_args()
    data_file = args.data_file

    vera_out_file = VeraOutFile(data_file)

    f = partial(_reload, vera_out_file=vera_out_file)

    # Make UI auto reload
    server.controller.on_server_reload.add(f)

    # Init application
    ui.initialize(server, vera_out_file)

    # Start server
    server.start(**kwargs)


if __name__ == "__main__":
    main()



from pathlib import Path

# Compute local path to serve
serve_path = str(Path(__file__).with_name("serve").resolve())

# Serve directory for JS/CSS files
serve = {"__vera_core": serve_path}

# List of JS files to load (usually from the serve path above)
scripts = ["__vera_core/vue-vera_core.umd.min.js"]

# List of CSS files to load (usually from the serve path above)
styles = ["__vera_core/vue-vera_core.css"]

# List of Vue plugins to install/load
vue_use = ["vera_core"]



import numpy as np
import plotly.express as px
import plotly.graph_objects as go

from trame.ui.html import DivLayout
from trame.widgets import plotly


OPTION = {
    "name": "time_plot",
    "label": "Time Plot",
    "icon": "mdi-chart-line",
}


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    def create_line(selected_array, indices=(0, 0, 0, 0)):
        exposures = [x.exposure[0] for x in vera_out_file.states]

        if selected_array == "pin_volumes":
            # It's just going to be a flat line. The volumes don't change.
            pin_volumes = vera_out_file.core.pin_volumes
            array = [pin_volumes[indices] for _ in vera_out_file.states]
        else:
            array = [getattr(x, selected_array)[indices] for x in vera_out_file.states]

        kwargs = {
            "x": exposures,
            "y": array,
            "labels": {
                "x": "exposure",
                "y": selected_array,
            },
        }
        figure = px.line(**kwargs)

        # Add a vertical line indicating our current exposure
        # # FIXME: why is add_vline only plotting from y==0 to y==1?
        # kwargs = {
        #     "x": vera_out_file.active_state.exposure[0],
        #     "line_dash": "dash",
        #     "line_color": "red",
        # }
        # figure.add_vline(**kwargs)
        #
        # # Because the above won't work, we have to make it manually
        float_info = np.finfo(np.float64)
        kwargs = {
            "x": [vera_out_file.active_state.exposure[0]] * 2,
            "y": [float_info.min, float_info.max],
            "mode": "lines",
            "line": go.scatter.Line(color="red", dash="dash"),
            "showlegend": False,
        }
        figure.add_trace(go.Scatter(**kwargs))

        figure.update_layout(margin=dict(t=0, b=0, l=0, r=0))
        return figure

    @state.change(
        "selected_array",
        "selected_assembly",
        "selected_layer",
        "selected_i",
        "selected_j",
    )
    @ctrl.add("on_vera_out_active_state_index_changed")
    def on_cell_change(
        selected_array,
        selected_assembly,
        selected_layer,
        selected_i,
        selected_j,
        **kwargs
    ):
        indices = (selected_j, selected_i, selected_layer, selected_assembly)
        indices = tuple(map(int, indices))
        ctrl.update_time_plot(create_line(selected_array, indices))

    with DivLayout(server, template_name="time_plot") as layout:
        layout.root.style = "height: 100%; width: 100%;"

        style = "; ".join(
            [
                "width: 100%",
                "height: 100%",
                "user-select: none",
            ]
        )
        figure = plotly.Figure(
            display_logo=False,
            display_mode_bar=False,
            style=style,
            # selected=(on_event, "["selected", utils.safe($event)]"),
            # hover=(on_event, "["hover", utils.safe($event)]"),
            # selecting=(on_event, "["selecting", $event]"),
            # unhover=(on_event, "["unhover", $event]"),
        )
        ctrl.update_time_plot = figure.update



import copy

import numpy as np

from vtkmodules.vtkCommonDataModel import vtkImageData, vtkPiecewiseFunction
from vtkmodules.vtkInteractionStyle import vtkInteractorStyleTrackballCamera
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor
from vtkmodules.vtkRenderingCore import (
    vtkColorTransferFunction,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
    vtkVolume,
    vtkVolumeProperty,
)
from vtkmodules.vtkRenderingVolumeOpenGL2 import vtkSmartVolumeMapper

import vtk.util.numpy_support as np_s

from trame.ui.html import DivLayout
from trame.widgets import vtk, vuetify

OPTION = {
    "name": "volume_view",
    "label": "Volume View",
    "icon": "mdi-rotate-3d",
}

reset_camera_count = 0


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    # Set up the VTK volume
    ren = vtkRenderer()
    ren_win = vtkRenderWindow()
    ren_win.AddRenderer(ren)
    ren_win.OffScreenRenderingOn()

    ren.SetBackground(1, 1, 1)

    iren = vtkRenderWindowInteractor()
    iren.SetInteractorStyle(vtkInteractorStyleTrackballCamera())
    iren.SetRenderWindow(ren_win)

    axes = vtkAxesActor()
    orientation_marker = vtkOrientationMarkerWidget()
    orientation_marker.SetOrientationMarker(axes)
    orientation_marker.SetInteractor(iren)
    # FIXME: I'm not sure if this is accurate, so it is disabled for now
    # orientation_marker.EnabledOn()
    # orientation_marker.InteractiveOn()

    # FIXME: for now, let's make it fully opaque so it matches veraview
    # exactly.
    opacity_points = [
        (0.00, 0),
        (0.01, 1),
        (1.95, 1),
    ]

    # Create transfer mapping scalar value to opacity.
    opacity_transfer_function = vtkPiecewiseFunction()
    for point in opacity_points:
        opacity_transfer_function.AddPoint(*point)

    # This is for a rainbow color map
    original_color_points = [
        (0.000000, 0.0, 0.0, 0.5625),
        (0.216992, 0.0, 0.0, 1.0000),
        (0.712975, 0.0, 1.0, 1.0000),
        (0.960965, 0.5, 1.0, 0.5000),
        (1.208960, 1.0, 1.0, 0.0000),
        (1.704940, 1.0, 0.0, 0.0000),
        (1.952930, 0.5, 0.0, 0.0000),
    ]

    # Create transfer mapping scalar value to color.
    color_transfer_function = vtkColorTransferFunction()
    for point in original_color_points:
        color_transfer_function.AddRGBPoint(*point)

    # The property describes how the data will look.
    volume_property = vtkVolumeProperty()
    volume_property.SetColor(color_transfer_function)
    volume_property.SetScalarOpacity(opacity_transfer_function)
    # I don't think we want to shade the data or perform linear interpolation
    # volume_property.ShadeOn()
    # volume_property.SetInterpolationTypeToLinear()

    volume_data = vtkImageData()
    volume_mapper = vtkSmartVolumeMapper()
    volume_mapper.SetInputData(volume_data)

    volume = vtkVolume()
    volume.SetMapper(volume_mapper)
    volume.SetProperty(volume_property)

    ren.AddVolume(volume)

    # Pitch the camera by 90 degrees to start
    ren.GetActiveCamera().Pitch(90)
    ren.GetActiveCamera().OrthogonalizeViewUp()

    @state.change("color_range")
    def update_color_points(color_range, **kwargs):
        # Rescale our color points when the color range changes
        original_range = (original_color_points[0][0], original_color_points[-1][0])
        new_color_points = copy.deepcopy(original_color_points)
        for i, row in enumerate(new_color_points):
            value = row[0]
            new_value = np.interp(value, original_range, color_range)
            new_color_points[i] = (new_value, *row[1:])

        color_transfer_function.RemoveAllPoints()
        for point in new_color_points:
            color_transfer_function.AddRGBPoint(*point)

        ren_win.Render()
        ctrl.view_update()

    @state.change("selected_array")
    @ctrl.add("on_vera_out_active_state_index_changed")
    def update_volume_view(selected_array, **kwargs):
        global reset_camera_count
        array = vera_out_file.array(selected_array)

        # Let's convert the data into a volume format
        assembly_shape = array.shape[:2]
        reduced_core_map = vera_out_file.core.reduced_core_map
        reduced_map_shape = reduced_core_map.shape
        expanded_core_shape = (
            reduced_map_shape[0] * assembly_shape[0],
            reduced_map_shape[1] * assembly_shape[1],
        )
        volume_shape = (*expanded_core_shape, array.shape[2])
        volume_array = np.zeros(volume_shape, dtype=array.dtype)

        # Now that we have the array, let's copy the assemblies into it
        for assembly_id in range(array.shape[3]):
            core_i, core_j = map(int, np.where(reduced_core_map == assembly_id + 1))
            i_range = (core_i * assembly_shape[0], (core_i + 1) * assembly_shape[0])
            j_range = (core_j * assembly_shape[1], (core_j + 1) * assembly_shape[1])

            volume_array[slice(*i_range), slice(*j_range)] = array[:, :, :, assembly_id]

        # It's possible that a rectilinear grid would be better here rather
        # than repeating voxels. But it also doesn't look super straightforward
        # to render a rectilinear grid as a volume. So let's just repeat the
        # voxels instead.
        axial_mesh_pixels = vera_out_file.core.axial_mesh_pixels
        volume_array = np.repeat(volume_array, axial_mesh_pixels, axis=2)

        # Since VTK uses Fortran ordering, we should transpose before raveling.
        raveled = volume_array.transpose(2, 1, 0).ravel()
        vtk_array = np_s.numpy_to_vtk(raveled, deep=True)

        volume_data.SetDimensions(*volume_array.shape)
        pd = volume_data.GetPointData()

        # Remove all other arrays
        while pd.GetNumberOfArrays() > 0:
            pd.RemoveArray(0)

        # Add the array
        pd.SetScalars(vtk_array)
        volume_data.Modified()

        # Update the view
        # ren_win.Render()
        if reset_camera_count < 2:
            ren.ResetCameraClippingRange()
            ren.ResetCamera()
            reset_camera_count += 1
            ctrl.reset_camera()

        ctrl.view_update()

        # Save a copy of the orientation marker so it won't go out of scope
        om = orientation_marker  # noqa

    with DivLayout(server, template_name="volume_view") as layout:
        layout.root.style = "height: 100%;"
        html_view = vtk.VtkRemoteView(ren_win, ref="volume_view")
        ctrl.reset_camera = html_view.reset_camera
        ctrl.view_update = html_view.update

        with vuetify.VBtn(
            icon=True,
            click=html_view.reset_camera,
            absolute=True,
            style="top: 0; right: 0",
            small=True,
        ):
            vuetify.VIcon("mdi-crop-free", small=True)



import numpy as np

from trame.ui.html import DivLayout
from vera_core.widgets import vera

OPTION = {
    "name": "assembly_view",
    "label": "Assembly View",
    "icon": "mdi-dots-grid",
}


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    # A cache of assembly images.
    cached_assembly_images = {}

    @state.change(
        "assembly_view_size",
        "selected_array",
        "selected_assembly",
        "selected_layer",
        "color_range",
    )
    @ctrl.add("on_vera_out_active_state_index_changed")
    def update_assembly_view(
        selected_time, selected_array, selected_assembly, selected_layer, **kwargs
    ):
        selected_assembly = int(selected_assembly)
        selected_layer = int(selected_layer)
        image_data = None

        # Extract from cache if possible
        cache_key = (selected_time, selected_array, selected_assembly, selected_layer)
        if cache_key in cached_assembly_images:
            # Shortcut if we have a cache. We might still need to redraw
            # if the figure size was updated.
            image_data = cached_assembly_images[cache_key]

        # Extract data from H5 + add to cache
        if image_data is None:
            array = vera_out_file.array(selected_array)
            image_data = array[:, :, selected_layer, selected_assembly]
            control_rod_positions = vera_out_file.core.control_rod_positions

            # Make control rod positions equal to nan
            image_data[control_rod_positions] = np.nan

            # Only allow one image in the cache
            MAX_ITEMS_IN_CACHE = 1
            while len(cached_assembly_images) >= MAX_ITEMS_IN_CACHE:
                cached_assembly_images.pop(next(iter(cached_assembly_images)))

            cached_assembly_images[cache_key] = image_data

        # Update the client
        state.assembly_array = np.ravel(image_data).tolist()

    # UI content
    with DivLayout(server, template_name="assembly_view") as layout:
        layout.root.style = "height: 100%;"
        vera.AssemblyView(
            value=("assembly_array", []),
            selected_i=("selected_i", 7),
            selected_j=("selected_j", 7),
            color_preset="jet",
            color_range=("color_range", [0, 3]),
            click="setAll({ selected_i: $event.i, selected_j: $event.j})",
        )



from trame.ui.html import DivLayout
from trame.widgets import vuetify


OPTION = {
    "name": "table_view",
    "label": "Table View",
    "icon": "mdi-table",
}


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    @state.change(
        "selected_array",
        "selected_assembly",
        "selected_layer",
        "selected_i",
        "selected_j",
    )
    @ctrl.add("on_vera_out_active_state_index_changed")
    def update_table(
        selected_array,
        selected_assembly,
        selected_layer,
        selected_i,
        selected_j,
        **kwargs,
    ):
        indices = (selected_j, selected_i, selected_layer, selected_assembly)
        indices = tuple(map(int, indices))

        array = vera_out_file.array(selected_array)
        value = array[indices]

        data_dict = {
            selected_array: value,
            "crit_boron": vera_out_file.active_state.crit_boron[0],
            "exposure": vera_out_file.active_state.exposure[0],
            "keff": vera_out_file.active_state.keff[0],
        }

        # Round floats so we don't display too many sig figs.
        # 7 sig figs matches veraview.
        sig_figs = 7
        data_dict = {
            k: float(f"{v:0.{sig_figs}g}")
            for k, v in data_dict.items()
            if isinstance(v, float)
        }

        # Create column labels
        axial_value = vera_out_file.core.axial_mesh_means[selected_layer]
        assembly_label = vera_out_file.core.reduced_core_map_label(selected_assembly)
        columns = [
            "Dataset",
            f"Assembly {assembly_label}; Axial {axial_value:0.6g} cm; Pin ({selected_i + 1}, {selected_j + 1})",
        ]

        # Convert into VDataTable format
        headers = [{"text": x, "value": x} for x in columns]

        data = list(data_dict.items())
        rows = [dict(zip(columns, entry)) for entry in data]

        state.table_view_headers = headers
        state.table_view_rows = rows

    table_options = {
        "headers": ("table_view_headers", []),
        "items": ("table_view_rows", []),
        "classes": "mt-n2",
        "disable_sort": True,
        "dense": True,
        "disable_pagination": True,
        # Even though the pagination is disabled, this is necessary to hide it
        "hide_default_footer": True,
    }

    with DivLayout(server, template_name="table_view") as layout:
        layout.root.style = "height: 100%; width: 100%;"
        vuetify.VDataTable(**table_options)



from typing import Union

import h5py
import numpy as np
import string

H5_ARRAY_TYPE = Union[h5py.Dataset, np.ndarray]


class VeraOutFile:
    def __init__(self, filename):
        # Keep this open for better performance
        self.f = h5py.File(filename, "r")
        self.core = VeraOutCore(self.f)
        self.core._cache_all()

        self.states = []

        self._create_states()

        self.active_state_index = 0

    def close(self):
        self.f.close()

    def _create_states(self):
        state_keys = [key for key in self.f if key.startswith("STATE_")]
        indices = [int(key.split("_")[1]) for key in state_keys]
        self.states = [VeraOutState(self.f, idx) for idx in indices]

    @property
    def active_state(self):
        return self.states[self.active_state_index]

    @property
    def active_state_index(self):
        return self._active_state_index

    @active_state_index.setter
    def active_state_index(self, index):
        if hasattr(self, "_active_state_index"):
            if self._active_state_index == index:
                return
            else:
                # Clear the cache from the active state
                self.active_state._uncache_all()

        self._active_state_index = index
        self.active_state._cache_all()

    def array(self, array_name):
        # Get the array with the name "array_name", either on the active state,
        # or on the core.

        # These are on the core
        arrays_on_core = [
            "pin_volumes",
        ]
        if array_name in arrays_on_core:
            # This one is on the core
            return getattr(self.core, array_name)

        # If not on the core, assume it is on the active states.
        return getattr(self.active_state, array_name)


class LazyHDF5Loader:
    def __init__(self, f, path, dataset_names):
        self._f = f
        self._path = path
        self._dataset_names = dataset_names

        self._uncache_all()

    def _load_dataset(self, name):
        return self._f[f"{self._path}/{name}"]

    def _cache(self, name):
        # Set the attribute to be the loaded numpy array
        if name not in self._dataset_names:
            raise AttributeError(name)

        setattr(self, name, self._load_dataset(name)[:])

    def _uncache(self, name):
        # Set the attribute to be the h5py dataset
        if name not in self._dataset_names:
            raise AttributeError(name)

        setattr(self, name, self._load_dataset(name))

    def _cache_all(self):
        for name in self._dataset_names:
            self._cache(name)

    def _uncache_all(self):
        for name in self._dataset_names:
            self._uncache(name)


class VeraOutCore(LazyHDF5Loader):
    # These are the attributes that will be read from the HDF5 file
    axial_mesh: H5_ARRAY_TYPE = None
    core_map: H5_ARRAY_TYPE = None
    core_sym: H5_ARRAY_TYPE = None
    pin_volumes: H5_ARRAY_TYPE = None

    def __init__(self, f):
        super().__init__(f, "/CORE", list(self.__annotations__))
        self.compute_reduced_core_map()
        self.compute_axial_mesh_pixels()
        self.compute_control_rod_positions()
        self.compute_axial_mesh_means()

    def compute_reduced_core_map(self):
        """Compute the reduced core map based upon the core_sym"""
        sym = self.core_sym[0]
        if sym == 1:
            self.reduced_core_map = self.core_map[:].copy()
            self.reduced_core_map_start_index = 0
        elif sym == 4:
            w, h = self.core_map[:].shape
            start_w = w // 2
            start_h = h // 2
            self.reduced_core_map = self.core_map[start_w:, start_h:]
            self.reduced_core_map_start_index = start_w
        else:
            raise Exception(f"Unhandled symmetry: {sym}")

        num_cols = self.reduced_core_map.shape[1]
        alphabet = [*string.ascii_uppercase]
        self.reduced_core_map_column_labels = list(reversed(alphabet[:num_cols]))

    def compute_axial_mesh_pixels(self):
        """Compute the number of pixels that we will be displaying in
        the axial direction for each length in the axial mesh.
        """
        diff_array = np.diff(self.axial_mesh[:])

        # The min diff will be three pixels high. The rest will be computed based
        # upon the min diff.
        MIN_DIFF_PIXELS_HEIGHT = 3
        pixel_height = np.min(diff_array) / MIN_DIFF_PIXELS_HEIGHT
        pixel_height_array = diff_array / pixel_height
        self.axial_mesh_pixels = np.round(pixel_height_array).astype(np.int64)

    def compute_control_rod_positions(self):
        # Assume they are the same in every volume
        first_volume = self.pin_volumes[:, :, 0, 0]
        self.control_rod_positions = np.where(first_volume == 0)

    def compute_axial_mesh_means(self):
        # Compute the mean between each neighbor
        repeats = [2] * len(self.axial_mesh)
        repeats[0] = 1
        repeats[-1] = 1

        repeated_mesh = np.repeat(self.axial_mesh, repeats)
        reshaped = repeated_mesh.reshape((repeated_mesh.shape[0] // 2, 2))

        self.axial_mesh_means = np.mean(reshaped, axis=1)

    def row_assembly_indices(self, assembly_idx):
        """Get indices of all assemblies in the same row as this assembly"""
        # The core map and reduced core map use 1-based indexing
        row = np.where(self.reduced_core_map == assembly_idx + 1)[0][0]
        ids = self.reduced_core_map[row]
        # Remove any zeros
        ids = ids[ids > 0]
        return ids - 1

    def col_assembly_indices(self, assembly_idx):
        """Get indices of all assemblies in the same column as this assembly"""
        col = np.where(self.reduced_core_map == assembly_idx + 1)[1][0]
        ids = self.reduced_core_map[:, col]
        # Remove any zeros
        ids = ids[ids > 0]
        return ids - 1

    def reduced_core_map_assembly(self, i, j):
        # Get the index of the assembly at reduced core map position i, j
        return int(self.reduced_core_map[j, i] - 1)

    def reduced_core_map_ij(self, assembly_idx):
        j, i = map(int, np.where(self.reduced_core_map == assembly_idx + 1))
        return i, j

    def reduced_core_map_label(self, assembly_idx):
        row_label = self.reduced_core_map_row_label(assembly_idx)
        col_label = self.reduced_core_map_column_label(assembly_idx)
        return f"{col_label}-{row_label}"

    def reduced_core_map_row_label(self, assembly_idx):
        i, j = self.reduced_core_map_ij(assembly_idx)
        start_index = self.reduced_core_map_start_index
        rows = list(range(start_index + 1, len(self.core_map) + 1))
        return str(rows[j])

    def reduced_core_map_column_label(self, assembly_idx):
        i, j = self.reduced_core_map_ij(assembly_idx)
        return self.reduced_core_map_column_labels[i]


class VeraOutState(LazyHDF5Loader):
    # These are the attributes that will be read from the HDF5 file
    crit_boron: H5_ARRAY_TYPE = None
    detector_response: H5_ARRAY_TYPE = None
    exposure: H5_ARRAY_TYPE = None
    keff: H5_ARRAY_TYPE = None
    pin_cladtemps: H5_ARRAY_TYPE = None
    pin_fueltemps: H5_ARRAY_TYPE = None
    pin_moddens: H5_ARRAY_TYPE = None
    pin_modtemps: H5_ARRAY_TYPE = None
    pin_powers: H5_ARRAY_TYPE = None

    def __init__(self, f, idx):
        super().__init__(f, f"/STATE_{idx:04}", list(self.__annotations__))
        self._index = idx



import numpy as np

from trame.ui.html import DivLayout
from vera_core.widgets import vera

OPTION = {
    "name": "y_axial_view",
    "label": "Y Axial View",
    "icon": "mdi-border-vertical",
}


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    def axial_cell_selected(layer, assembly_j):
        assembly_i = state.selected_assembly_ij["i"]
        state.selected_assembly = vera_out_file.core.reduced_core_map_assembly(
            assembly_i, assembly_j
        )
        state.selected_layer = layer

    @state.change(
        "selected_array",
        "selected_assembly",
        "selected_i",
    )
    @ctrl.add("on_vera_out_active_state_index_changed")
    def update_axial_view(selected_array, selected_assembly, selected_i, **kwargs):
        selected_assembly = int(selected_assembly)
        selected_i = int(selected_i)
        col_assembly_indices = vera_out_file.core.col_assembly_indices(
            selected_assembly
        )
        array = vera_out_file.array(selected_array)
        assembly_size = array.shape[0]

        # Numpy will tack the indexing subspace on to the beginning
        image_data = array[:, selected_i, :, col_assembly_indices]
        image_data = np.vstack(image_data).T

        # Have to reverse the y-axis since ax.invert_yaxis() doesn't work here
        image_data = image_data[::-1, :]

        # Extract cell sizes
        nb_lines = image_data.shape[0]
        nb_cols = int(image_data.shape[1] / assembly_size)
        state.y_axial_core_size_y = vera_out_file.core.axial_mesh_pixels.tolist()
        state.y_axial_core_size_x = [assembly_size for i in range(nb_cols)]
        state.y_axial_core_label_y = [
            i + 1 for i in range(len(state.y_axial_core_size_y))
        ]
        state.y_axial_core_label_y.reverse()

        # Update UI
        state.y_axial_core = []
        for j in range(nb_lines):
            line = []
            state.y_axial_core.append(line)
            for i in range(nb_cols):
                assembly = image_data[
                    j, slice(i * assembly_size, (i + 1) * assembly_size)
                ]
                line.append(np.ravel(assembly).tolist())

    # Convert these to 1-based indexing
    start_x = vera_out_file.core.reduced_core_map_start_index + 1
    stop_x = len(vera_out_file.core.core_map) + 1

    with DivLayout(server, template_name="y_axial_view") as layout:
        layout.root.style = "height: 100%;"
        vera.AxialView(
            value=("y_axial_core", []),
            color_preset="jet",
            color_range=("color_range", [0, 3]),
            x_sizes=("y_axial_core_size_x", []),
            y_sizes=("y_axial_core_size_y", []),
            x_labels=("y_axial_core_label_x", list(range(start_x, stop_x))),
            y_labels=("y_axial_core_label_y", []),
            selected_i=("selected_assembly_ij.j",),
            selected_j=("y_axial_core_label_y.length - selected_layer - 1",),
            click=(
                axial_cell_selected,
                "[y_axial_core_label_y.length - $event.j - 1, $event.i]",
            ),
            x_scale=("3",),
            y_scale=("3",),
        )



import numpy as np

from trame.ui.vuetify import SinglePageLayout
from trame.widgets import vuetify, grid, client, html
from vera_core.widgets import vera
from . import (
    assembly_view,
    axial_plot,
    core_view,
    empty,
    table_view,
    time_plot,
    x_axial_view,
    y_axial_view,
    volume_view,
    assets,
)

DEFAULT_NB_ROWS = 8


def get_next_y_from_layout(layout):
    next_y = 0
    for item in layout:
        y, h = item.get("y", 0), item.get("h", 1)
        if y + h > next_y:
            next_y = y + h
    return next_y


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller
    state.trame__title = "VERACore"

    # FIXME: For our example, fix this to match VeraView.
    # Come up with a way to autogenerate it.
    state.color_range = (0.0273, 1.95)
    state.selected_layer = 24
    state.selected_assembly = 36
    state.selected_time = 0
    # FIXME ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    @state.change("selected_time")
    def selected_time_changed(selected_time, **kwargs):
        selected_time = int(selected_time)
        vera_out_file.active_state_index = selected_time

        ctrl.on_vera_out_active_state_index_changed(
            selected_time=selected_time, **kwargs
        )

    @state.change("selected_array")
    def selected_array_changed(selected_array, **kwargs):
        array = vera_out_file.array(selected_array)
        state.color_range = (np.nanmin(array), np.nanmax(array))

    # Keep selected_assembly and selected_assembly_ij in sync
    @state.change("selected_assembly_ij")
    def selected_assembly_ij_changed(selected_assembly_ij, **kwargs):
        i, j = selected_assembly_ij["i"], selected_assembly_ij["j"]
        state.selected_assembly = vera_out_file.core.reduced_core_map_assembly(i, j)

    @state.change("selected_assembly")
    def selected_assembly_changed(selected_assembly, **kwargs):
        i, j = vera_out_file.core.reduced_core_map_ij(selected_assembly)
        state.selected_assembly_ij = dict(i=i, j=j)

    # Go ahead and make sure they are in sync
    selected_assembly_changed(state.selected_assembly)

    # Initialize all visualizations
    state.setdefault("grid_options", [])
    state.setdefault("grid_layout", [])
    assembly_view.initialize(server, vera_out_file)
    axial_plot.initialize(server, vera_out_file)
    x_axial_view.initialize(server, vera_out_file)
    y_axial_view.initialize(server, vera_out_file)
    core_view.initialize(server, vera_out_file)
    table_view.initialize(server, vera_out_file)
    time_plot.initialize(server, vera_out_file)
    volume_view.initialize(server, vera_out_file)
    empty.initialize(server)

    # Reserve the various views
    available_view_ids = [f"{v+1}" for v in range(10)]
    for view_id in available_view_ids:
        state[f"grid_view_{view_id}"] = empty.OPTION

    # X Axial view
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=0, y=0, w=3, h=17, i=view_id),
    )
    state[f"grid_view_{view_id}"] = x_axial_view.OPTION

    # Y Axial view
    # view_id = available_view_ids.pop(0)
    # state.grid_layout.append(
    #     dict(x=3, y=0, w=3, h=17, i=view_id),
    # )
    # state[f"grid_view_{view_id}"] = y_axial_view.OPTION

    # Core view
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=6, y=0, w=3, h=9, i=view_id),
    )
    state[f"grid_view_{view_id}"] = core_view.OPTION

    # Assembly view
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=9, y=0, w=3, h=9, i=view_id),
    )
    state[f"grid_view_{view_id}"] = assembly_view.OPTION

    # Axial plot
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=6, y=9, w=3, h=8, i=view_id),
    )
    state[f"grid_view_{view_id}"] = axial_plot.OPTION

    # Time plot
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=9, y=9, w=3, h=8, i=view_id),
    )
    state[f"grid_view_{view_id}"] = time_plot.OPTION

    # Volume view
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=3, y=0, w=3, h=17, i=view_id),
    )
    state[f"grid_view_{view_id}"] = volume_view.OPTION

    # Table view
    view_id = available_view_ids.pop(0)
    state.grid_layout.append(
        dict(x=0, y=17, w=6, h=10, i=view_id),
    )
    state[f"grid_view_{view_id}"] = table_view.OPTION

    @ctrl.set("grid_add_view")
    def add_view():
        next_view_id = available_view_ids.pop()
        next_y = get_next_y_from_layout(state.grid_layout)
        state.grid_layout.append(
            dict(x=0, w=12, h=DEFAULT_NB_ROWS, y=next_y, i=next_view_id)
        )
        state.dirty("grid_layout")

    @ctrl.set("grid_remove_view")
    def remove_view(view_id):
        available_view_ids.append(view_id)
        state.grid_layout = list(
            filter(lambda item: item.get("i") != view_id, state.grid_layout)
        )

    # Setup main layout
    with SinglePageLayout(server) as layout:
        # Toolbar
        with layout.toolbar as toolbar:
            toolbar.clear()
            toolbar.height = 36

            html.Img(src=assets.LOGO, height=25)
            vuetify.VSpacer()

            vera.ColorMapEditor(
                v_model="color_range",
                color_preset="jet",
            )

            vuetify.VSpacer()

            vuetify.VSelect(
                v_model=("selected_array", "pin_powers"),
                items=(
                    "available_arrays",
                    [
                        dict(text="Pin Powers", value="pin_powers"),
                        dict(text="Pin Clad Temps", value="pin_cladtemps"),
                        dict(text="Pin Fuel Temps", value="pin_fueltemps"),
                        dict(text="Pin Moderator Density", value="pin_moddens"),
                        dict(text="Pin Moderator Temps", value="pin_modtemps"),
                        dict(text="Pin Volumes", value="pin_volumes"),
                    ],
                ),
                hide_details=True,
                dense=True,
                style="max-width: 220px",
            )

            with vuetify.VBtn(icon=True, click=ctrl.grid_add_view):
                vuetify.VIcon("mdi-plus")

        # Main content
        with layout.content:
            with vuetify.VContainer(fluid=True, classes="pa-0 fill-height"):
                with grid.GridLayout(
                    layout=("grid_layout", []),
                    row_height=30,
                    vertical_compact=True,
                    style="width: 100%; height: 100%;",
                ):
                    with grid.GridItem(
                        v_for="item in grid_layout",
                        key="item.i",
                        v_bind="item",
                        style="touch-action: none;",
                        drag_ignore_from=".drag_ignore",
                    ):
                        with vuetify.VCard(style="height: 100%;"):
                            with vuetify.VCardTitle(classes="py-1 px-1"):
                                with vuetify.VMenu(offset_y=True):
                                    with vuetify.Template(
                                        v_slot_activator="{ on, attrs }"
                                    ):
                                        with vuetify.VBtn(
                                            icon=True,
                                            small=True,
                                            v_bind="attrs",
                                            v_on="on",
                                        ):
                                            vuetify.VIcon(
                                                v_text="get(`grid_view_${item.i}`).icon"
                                            )
                                        html.Div(
                                            "{{ get(`grid_view_${item.i}`).label }}",
                                            classes="ml-1 text-subtitle-2",
                                        )
                                    with vuetify.VList(dense=True):
                                        with vuetify.VListItem(
                                            v_for="(option, index) in grid_options",
                                            key="index",
                                            click="set(`grid_view_${item.i}`, option)",
                                        ):
                                            with vuetify.VListItemIcon():
                                                vuetify.VIcon(v_text="option.icon")
                                            vuetify.VListItemTitle("{{ option.label }}")
                                vuetify.VSpacer()
                                with vuetify.VBtn(
                                    icon=True,
                                    x_small=True,
                                    click=(ctrl.grid_remove_view, "[item.i]"),
                                ):
                                    vuetify.VIcon(
                                        "mdi-delete-forever-outline", small=True
                                    )
                            vuetify.VDivider()

                            style = "; ".join(
                                [
                                    "position: relative",
                                    "height: calc(100% - 37px)",
                                    "overflow: auto",
                                ]
                            )
                            with vuetify.VCardText(style=style, classes="drag_ignore"):
                                # Add template for value of get(`grid_view_${item.i}`)
                                client.ServerTemplate(
                                    name=("get(`grid_view_${item.i}`).name",)
                                )

        # Footer
        with layout.footer as footer:
            footer.clear()
            footer.height = 35
            with vuetify.VBtn(
                icon=True,
                small=True,
                disabled=("selected_time == 0",),
                click="selected_time--",
            ):
                vuetify.VIcon("mdi-minus")
            html.Div(
                "State {{ selected_time }}",
                classes="text-center",
                style="width: 100px;",
            )
            with vuetify.VBtn(
                icon=True,
                small=True,
                disabled=(f"selected_time == {len(vera_out_file.states) - 1}",),
                click="selected_time++",
            ):
                vuetify.VIcon("mdi-plus")
            vuetify.VDivider(vertical=True, classes="mx-2")
            vuetify.VSlider(
                v_model=("selected_time",),
                min=0,
                max=len(vera_out_file.states) - 1,
                dense=True,
                hide_details=True,
                ticks="always",
                tick_size="4",
                height=35,
                # style="position: relative; top: -10px;",
                # tick_labels=(f"[{','.join([str(v) for v in range(len(vera_out_file.states))])}]", ),
            )



import numpy as np

from trame.ui.html import DivLayout
from vera_core.widgets import vera

OPTION = {
    "name": "core_view",
    "label": "Core View",
    "icon": "mdi-chart-pie",
}


def initialize(server, vera_out_file):
    state, ctrl = server.state, server.controller

    if OPTION not in state.grid_options:
        state.grid_options.append(OPTION)

    @state.change("selected_array", "selected_layer")
    @ctrl.add("on_vera_out_active_state_index_changed")
    def update_core_view(selected_array, selected_layer, **kwargs):
        selected_layer = int(selected_layer)

        array = vera_out_file.array(selected_array)

        # Load the layer and swap axes for faster indexing
        layer_array = array[:, :, selected_layer].swapaxes(0, 2)
        layer_array = layer_array.swapaxes(1, 2)

        control_rod_positions = vera_out_file.core.control_rod_positions
        reduced_core_map = vera_out_file.core.reduced_core_map
        core_width = reduced_core_map.shape[0]

        # custom widget update
        state.core_assemblies = []
        for i in range(core_width):
            line = []
            state.core_assemblies.append(line)
            for j in range(core_width):
                index = reduced_core_map[i, j] - 1
                if index == -1:
                    continue

                assembly_array = layer_array[index]

                # Set control rod positions to be nan
                assembly_array[control_rod_positions] = np.nan

                line.append(np.ravel(assembly_array).tolist())

    with DivLayout(server, template_name="core_view") as layout:
        layout.root.style = "height: 100%;"
        vera.CoreView(
            value=("core_assemblies", []),
            selected_i=("selected_assembly_ij.i",),
            selected_j=("selected_assembly_ij.j",),
            color_preset="jet",
            color_range=("color_range", [0, 3]),
            click="selected_assembly_ij = $event",
        )
